#include <iostream>
#include <Eigen/Core>
#include <random>
#include "toy_targets.h"

using namespace std;
using namespace Eigen;

const double _PI = 3.14159265358979323846;

// To test on its own, remove the comments around main() then
// g++ -I /usr/include/eigen3 toy_targets.cpp
// When linking with another main file: #include "toy_targets.h"

//----------------------------------

// Null functions
// For flat priors or flat likelihood
double l_null(const ArrayXd &x, const ArrayXd &thetas) {
  return 0.0;
}
VectorXd gl_null(const ArrayXd &x, const ArrayXd &thetas) {
  return x*0;
}
double gldote_null(const ArrayXd &x, const VectorXd &e, const ArrayXd &thetas) {
  return 0.0;
}

//**********ISOTROPIC****************

// Isotropic Gaussian
double l_GaussIso(const ArrayXd &x, const ArrayXd &thetas) {
  double norm=x.matrix().norm();
  double sig2=thetas[0];
  return -0.5*norm*norm/sig2;
}
VectorXd gl_GaussIso(const ArrayXd &x, const ArrayXd &thetas) {
  double sig2=thetas[0];
  return -x.matrix()/sig2;
}
double gldote_GaussIso(const ArrayXd &x, const VectorXd &e, const ArrayXd &thetas) {
  double sig2=thetas[0];
  return -e.dot(x.matrix())/sig2;
}

// Power isotropic log f = -||x||^a/a where ||x|| is the L2 norm
double l_PowIso(const ArrayXd &x, const ArrayXd &theta) {
  double norm=x.matrix().norm();
  double a=theta[0];
  return -exp(a*log(norm))/a;
}
VectorXd gl_PowIso(const ArrayXd &x, const ArrayXd &theta) {
  double norm=x.matrix().norm();
  double a=theta[0];
  return -exp((a-2)*log(norm))*x.matrix();
}
double gldote_PowIso(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  double norm=x.matrix().norm();
  double a=theta[0];
  return -exp((a-2)*log(norm))*e.dot(x.matrix());
}


// Power product log f = -(1/a)\sum |x_i/lambda_i|^a
double l_PowProd(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd sc=theta.tail(d);
  double a=theta[0];
  return -pow(x.abs()/sc,a).sum()/a;
}

VectorXd gl_PowProd(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd sc=theta.tail(d);
  double a=theta[0];
  return -(x.sign()/sc*pow(x.abs()/sc,a)).matrix();
}

double gldote_PowProd(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd sc=theta.tail(d);
  double a=theta[0];
  return -(x.sign()/sc*pow(x.abs()/sc,a)).matrix().dot(e);
}



//**********MORE GENERAL GAUSSIAN************

// Diagonal Gaussian specifying each expectation and diagonal variance.
// Expectation \ne 0 can be useful when using this as a prior.
double l_GaussDiag(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd E=theta.head(d), V=theta.tail(d);
  ArrayXd lterms=(x-E)*(x-E)/V;
  return -0.5*lterms.sum();
}
VectorXd gl_GaussDiag(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd E=theta.head(d), V=theta.tail(d);
  ArrayXd gterms=-(x-E)/V;
  return gterms.matrix();
}
double gldote_GaussDiag(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  return e.dot(gl_GaussDiag(x,theta));
}

// SDs linear from theta[0] to theta[1]
double l_GaussLinSD(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd sds=ArrayXd::LinSpaced(x.size(),theta[0],theta[1]);
  ArrayXd tmp=x*x/sds/sds;
  return -0.5*tmp.sum();
}
VectorXd gl_GaussLinSD(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd sds=ArrayXd::LinSpaced(x.size(),theta[0],theta[1]);
  ArrayXd tmp=-x/(sds*sds);
  return tmp.matrix();
}
double gldote_GaussLinSD(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  return e.dot(gl_GaussLinSD(x,theta));
}

// Diag(Hessian) linear from theta[0] to theta[1]
double l_GaussLinH(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd Hdiag=ArrayXd::LinSpaced(x.size(),theta[0],theta[1]);
  ArrayXd tmp=x*x*Hdiag;
  return -0.5*tmp.sum();
}
VectorXd gl_GaussLinH(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd Hdiag=ArrayXd::LinSpaced(x.size(),theta[0],theta[1]);
  ArrayXd tmp=-x*Hdiag;
  return tmp.matrix();
}
double gldote_GaussLinH(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  return e.dot(gl_GaussLinH(x,theta));
}


//*************** MORE GENERAL EXPONENTIAL POWER LAW **************

// Power aniso log f = -||x/scvec||^a/a theta=(a,lo,hi)
double l_PowLinSc(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double norm=xstd.matrix().norm();
  double a=theta[0];
  return -exp(a*log(norm))/a;
}
VectorXd gl_PowLinSc(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double norm=xstd.matrix().norm();
  double a=theta[0];
  ArrayXd g=-exp((a-2)*log(norm))*xstd/scs;
  return g.matrix();
}
double gldote_PowLinSc(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  return e.dot(gl_PowLinSc(x,theta));
}

// Product of independent powers: log f = -sum |x_i/sc_i|^a/a theta=(a,lo,hi)
double l_PowProdLinSc(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double a=theta[0];
  ArrayXd terms=exp(a*log(xstd.abs()));
  return -terms.sum()/a;
}
VectorXd gl_PowProdLinSc(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double am1=theta[0]-1;
  ArrayXd g=-exp(am1*log(xstd.abs()))*xstd.sign()/scs;
  return g.matrix();
}
double gldote_PowProdLinSc(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  return e.dot(gl_PowProdLinSc(x,theta));
}

//************** Logistic *******************
// Product of logistics along the co-ordinate axes
// vector specifies (E1,...,Ed,Sc1,...,Scd)
double l_LogisticCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd E=theta.head(d), Sc=theta.tail(d);
  ArrayXd std=(x-E)/Sc; // standardised
  double m=std.maxCoeff(); // for robustness to large std
  return (std-2*(m+log(exp(-m)+exp(std-m)))-log(Sc)).sum();
}
VectorXd gl_LogisticCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd E=theta.head(d), Sc=theta.tail(d);
  ArrayXd std=(x-E)/Sc;
  ArrayXd gterms=1/Sc * (1-2/(1+exp(-std)));
  return gterms.matrix();
}



//*************** Student-t ******************

// Diagonal multivariate Student-t with linear scalings
double l_MVtDiagLinSc(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double df=theta(0);
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double denom=1+(xstd*xstd).sum()/df;
  return -(df+(double)d)/2*log(denom);
}
VectorXd gl_MVtDiagLinSc(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double df=theta(0);
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double denom=1+(xstd*xstd).sum()/df;
  return -(df+(double)d)/df*xstd/scs/denom;
}
double gldote_MVtDiagLinSc(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  return e.dot(gl_MVtDiagLinSc(x,theta));
}

// Product of univariate Student-ts with linear scalings
double l_tProdLinSc(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double df=theta[0];
  ArrayXd terms=-(df+1)/2*log(1+xstd*xstd/df);
  return terms.sum();
}
VectorXd gl_tProdLinSc(const ArrayXd &x, const ArrayXd &theta) {
  ArrayXd scs=ArrayXd::LinSpaced(x.size(),theta[1],theta[2]);
  ArrayXd xstd=x/scs;
  double df=theta[0];
  ArrayXd terms=-(df+1)/df*xstd/scs/(1+xstd*xstd/df);
  return terms.matrix();
}
double gldote_tProdLinSc(const ArrayXd &x, const VectorXd &e, const ArrayXd &theta) {
  return e.dot(gl_tProdLinSc(x,theta));
}

// Multivariate Student-t with general inverse variance matrix
double l_MVt(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double df=theta(0);
  ArrayXd Vinvlong=theta.tail(d*d);
  Map<MatrixXd> Vinv(Vinvlong.data(),d,d);
  VectorXd xv=x.matrix();
  double denom=1+(xv.transpose()*Vinv*xv)(0,0)/df;
  return  - (df+(double)d)/2.0 *log(denom);
}

// *********Skew Normal*********
// **Ancillary functions**
double Phi(const double x) {
  const double sqrt2=sqrt(2.0);
  return 0.5*(1.0+std::erf(x/sqrt2));
}
double logPhi(const double x) {
  if (x>-7) {
    if (x>7) {
      return -exp(logPhi(-x)); // log(Phi(x))=log(1-Phi(-x))
    }
    else {
      return log(Phi(x));
    }
  }
  else { // x is negative, remember!
    double x2=x*x, x3=x2*x, x5=x3*x*x, x7=x5*x*x;
    return -0.5*log(2*_PI)-0.5*x*x+log(-1/x+1/x3-3/x5+15/x7);
  }
}
double phiOPhi(const double x) {
  if (x>-7) {
    return 1/sqrt(2*_PI)*exp(-x*x/2)/Phi(x);
  }
  else {
    double x3=x*x*x, x5=x3*x*x, x7=x5*x*x;
    return 1/(-1/x+1/x3-3/x5+15/x7);
  }
}
ArrayXd Phi(const ArrayXd &xs) {
  ArrayXd Phis=xs;
  int i;
  const int n=xs.size();
  const double sqrt2=sqrt(2.0);
  
  for (i=0; i<n; i++) {
    Phis(i)=0.5*(1+std::erf(xs(i)/sqrt2));
  }
  return Phis;
}
ArrayXd logPhi(const ArrayXd &xs) {
  ArrayXd logPhis=xs;
  int i;
  const int n=xs.size();

  for (i=0; i<n; i++) {
    logPhis(i)=logPhi(xs(i));
  }
  return logPhis;
}
ArrayXd phiOPhi(const ArrayXd &xs) {
  ArrayXd rats=xs;
  int i;
  const int n=xs.size();

  for (i=0; i<n; i++) {
    rats(i)=phiOPhi(xs(i));
  }
  return rats;
}
// Product of skey normals along the co-ordinate axes
// vector specifies (alpha,Sc1,...,Scd)
double l_SkewNormalCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double alph=theta(0);
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised
  return (log(2/Sc)-0.5*std*std-0.5*log(2*_PI)+logPhi(alph*std)).sum();
}
VectorXd gl_SkewNormalCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double alph=theta(0);
  ArrayXd Sc=theta.tail(d), std=x/Sc;
  ArrayXd gterms= -std/Sc + alph/Sc * phiOPhi(alph*std);
  return gterms.matrix();
}

// ******** Quartic x Gaussian product target from HnH *********
double l_QuartGaussCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double a=theta(0);
  ArrayXd Sc=theta.tail(d), std2=x*x/Sc/Sc; // standardised squared
  return -0.25*(std2*std2).sum()-0.5*std2.sum()/a/a;
}
VectorXd gl_QuartGaussCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double a=theta(0);
  ArrayXd Sc=theta.tail(d), std=x/Sc;
  ArrayXd gterms= -std*std*std/Sc -std/(a*a*Sc);
  return gterms.matrix();
}
// ******** Logistic x Gaussian product target from HnH *********
double l_LogisticGaussCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double a=theta(0);
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised squared
  return -2*(log(cosh(0.5*std))).sum()-0.5*(std*std).sum()/a/a;
}
VectorXd gl_LogisticGaussCpts(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double a=theta(0);
  ArrayXd Sc=theta.tail(d), std=x/Sc;
  ArrayXd gterms= -tanh(0.5*std)/Sc -std/(a*a*Sc);
  return gterms.matrix();
}
// ******** BananaTwo in first two dimensions *********
double l_BananaTwo(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double lam=theta(0), den=1-lam*lam;
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised squared
  // Discrepancy from Gaussian
  double Delta=0.5*std(1)*std(1)-0.5*pow(std(1)-lam*(std(0)*std(0)-1),2.0)/den;
  return Delta-0.5*(std*std).sum();
}
VectorXd gl_BananaTwo(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double lam=theta(0), den=1-lam*lam;
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised
  ArrayXd gterms= -std;
  gterms(0)+=2*lam*std(0)/den*(std(1)-lam*(std(0)*std(0)-1));
  gterms(1)+=-1/den*(std(1)-lam*(std(0)*std(0)-1))+std(1);
  gterms/=Sc;
  return gterms.matrix();
}

// ******** Bimodal in first two dimensions *********
double l_BimodalGauss(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double lam=theta(0);
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised
  double z0a=std(0)-sqrt(lam), z1a=std(1)-sqrt(lam);
  double z0b=std(0)+sqrt(lam), z1b=std(1)+sqrt(lam);
  double tmpa=exp(-0.5/(1-lam)*(z0a*z0a+z1a*z1a));
  double tmpb=exp(-0.5/(1-lam)*(z0b*z0b+z1b*z1b));
  double tmp1=0.5*std(0)*std(0)+0.5*std(1)*std(1)+log(tmpa+tmpb);
  return tmp1-0.5*(std*std).sum();
}
VectorXd gl_BimodalGauss(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  double lam=theta(0);
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised
  double z0a=std(0)-sqrt(lam), z1a=std(1)-sqrt(lam);
  double z0b=std(0)+sqrt(lam), z1b=std(1)+sqrt(lam);
  double tmpa=exp(-0.5/(1-lam)*(z0a*z0a+z1a*z1a));
  double tmpb=exp(-0.5/(1-lam)*(z0b*z0b+z1b*z1b));
  double add0=(-1/(1-lam)*z0a*tmpa-1/(1-lam)*z0b*tmpb)/(tmpa+tmpb);
  double add1=(-1/(1-lam)*z1a*tmpa-1/(1-lam)*z1b*tmpb)/(tmpa+tmpb);
  ArrayXd gterms= -std;
  gterms(0)+=add0+std(0); gterms(1)+=add1+std(1);
  gterms/=Sc;
  return gterms.matrix();
}

// ******** Plus Prism in first two dimensions *********
double l_PlusPrism(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised
  double lam=theta(0), s0=std(0), s1=std(1);
  double tmpa=exp(-0.5/(1-lam)*s0*s0-0.5/(1+lam)*s1*s1);
  double tmpb=exp(-0.5/(1+lam)*s0*s0-0.5/(1-lam)*s1*s1);
  double tmp1=log(tmpa+tmpb)+0.5*s0*s0+0.5*s1*s1;
  return tmp1-0.5*(std*std).sum();
}
VectorXd gl_PlusPrism(const ArrayXd &x, const ArrayXd &theta) {
  int d=x.size();
  ArrayXd Sc=theta.tail(d), std=x/Sc; // standardised
  double lam=theta(0), s0=std(0), s1=std(1);
  double tmpa=exp(-0.5/(1-lam)*s0*s0-0.5/(1+lam)*s1*s1);
  double tmpb=exp(-0.5/(1+lam)*s0*s0-0.5/(1-lam)*s1*s1);
  double add0=(-1/(1-lam)*s0*tmpa-1/(1+lam)*s0*tmpb)/(tmpa+tmpb);
  double add1=(-1/(1-lam)*s1*tmpb-1/(1+lam)*s1*tmpa)/(tmpa+tmpb);
  ArrayXd gterms= -std;
  gterms(0)+=s0+add0; gterms(1)+=s1+add1;
  gterms/=Sc;
  return gterms.matrix();
}


// ******NUMERICAL DIFFERENTIATION*******
// Can be used to check the actual differentiation or to create bespoke
// gradient functions that use the numerical derivative

VectorXd glp_from_lp(const ArrayXd &xs, const ArrayXd &theta, double (*plp_fn)(const ArrayXd &xs, const ArrayXd &thetas), const double eps) {
  int i, d=xs.size();
  VectorXd glp=VectorXd(d);
  ArrayXd ei=ArrayXd::Zero(d);
  for (i=0;i<d;i++) {
    ei(i)=eps;
    glp[i]=((*plp_fn)(xs+ei,theta)-(*plp_fn)(xs-ei,theta))/(2*eps);
    ei(i)=0;
  }
  return glp;
}

MatrixXd Hlp_from_lp(const ArrayXd &xs, const ArrayXd &theta, double (*plp_fn)(const ArrayXd &xs, const ArrayXd &theta), const double eps) {
  int i,j, d=xs.size();
  MatrixXd Hlp=MatrixXd(d,d);
  ArrayXd ei=ArrayXd::Zero(d), ej=ArrayXd::Zero(d);
  for (i=0;i<d;i++) {
    ei(i)=eps;
    for (j=0;j<d;j++) {
      ej(j)=eps;
      Hlp(i,j)=(*plp_fn)(xs+ei+ej,theta)+(*plp_fn)(xs-ei-ej,theta)
	-(*plp_fn)(xs-ei+ej,theta)-(*plp_fn)(xs+ei-ej,theta);
      Hlp(i,j)=Hlp(i,j)/(4*eps*eps);
      ej(j)=0.0;
    }
    ei(i)=0.0;
  }
  return Hlp;
}


// ************* Random samples ***********************

Eigen::ArrayXd StdNormalAXd(size_t n, std::mt19937 &gen, std::normal_distribution<double> StdNormal) {
  size_t i;
  Eigen::ArrayXd out=Eigen::ArrayXd(n);
  for (i=0;i<n;i++) {
    out(i)=StdNormal(gen);
  }
  return out;
}

Eigen::ArrayXd StdSkewNormalAXd(size_t n, double omega, std::mt19937 &gen, std::normal_distribution<double> StdNormal) {
  size_t i;
  Eigen::ArrayXd out=Eigen::ArrayXd(n);
  for (i=0;i<n;i++) {
    double tmp1=StdNormal(gen), tmp2=StdNormal(gen);
    if (tmp2<omega*tmp1) { // has probability Phi(omega tmp1)
      out(i)=tmp1;
    }
    else {
      out(i)=-tmp1;
    }
  }
  return out;
}

Eigen::ArrayXd StdUnifAXd(size_t n, std::mt19937 &gen, std::uniform_real_distribution<double> StdUnif) {
  size_t i;
  Eigen::ArrayXd out=Eigen::ArrayXd(n);
  for (i=0;i<n;i++) {
    out(i)=StdUnif(gen);
  }
  return out;
}

Eigen::ArrayXd StdLogisticAXd(size_t n, std::mt19937 &gen, std::uniform_real_distribution<double> StdUnif) {
  size_t i;
  Eigen::ArrayXd out=Eigen::ArrayXd(n);
  for (i=0;i<n;i++) {
    out(i)=StdUnif(gen);
  }
  return log(out/(1-out));
}

double Targd::l_fn(const Eigen::ArrayXd &xs) {
  return pl_fn(xs,theta);
}
Eigen::VectorXd Targd::gl_fn(const Eigen::ArrayXd &xs) {
  return pgl_fn(xs,theta);
}



/*

// Test 
int main() {
  int d=4,i,j;
  ArrayXd x=ArrayXd::LinSpaced(d,1,d);
  double l,a;
  VectorXd gl;
  VectorXd e=VectorXd::Zero(d); e(0)=1;
  
  cout << "x="<<x.matrix().transpose()<<endl;

  cout << "Null functions\n";
  ArrayXd thetanull(1); thetanull << 1;
  l=l_null(x,thetanull);
  gl=gl_null(x,thetanull);
  a=gldote_null(x,e,thetanull);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;

  ArrayXd thetaGaussIso(1); thetaGaussIso <<2;
  cout << "Isotropic Gaussian with sigma2="<<thetaGaussIso[0]<<endl;
  l=l_GaussIso(x,thetaGaussIso);
  gl=gl_GaussIso(x,thetaGaussIso);
  a=gldote_GaussIso(x,e,thetaGaussIso);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaGaussIso, &l_GaussIso);
  cout << gl.transpose()<<endl;

  ArrayXd thetaPowIso(1); thetaPowIso <<4; // is the power
  cout << "Isotropic L2 norm to power of "<<thetaPowIso[0]<<endl;
  l=l_PowIso(x,thetaPowIso);
  gl=gl_PowIso(x,thetaPowIso);
  a=gldote_PowIso(x,e,thetaPowIso);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaPowIso, &l_PowIso);
  cout << gl.transpose()<<endl;

  ArrayXd thetaPowProd(d+1); thetaPowProd(0)=4; // is the power
  thetaPowProd.tail(d)=ArrayXd::LinSpaced(d,1,d);
  cout << "Product of abs cpts power of "<<thetaPowProd[0]<<endl;
  l=l_PowProd(x,thetaPowProd);
  gl=gl_PowProd(x,thetaPowProd);
  a=gldote_PowProd(x,e,thetaPowProd);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaPowProd, &l_PowProd);
  cout << gl.transpose()<<endl;
  
  ArrayXd thetaGaussDiag(2*d);
  thetaGaussDiag.head(d)=ArrayXd::LinSpaced(d,0,d-1);
  thetaGaussDiag.tail(d)=ArrayXd::LinSpaced(d,1,d);
  cout << "Gaussian expectations and diagonal variances: ";
  cout << thetaGaussDiag.matrix().transpose()<<endl;
  l=l_GaussDiag(x,thetaGaussDiag);
  gl=gl_GaussDiag(x,thetaGaussDiag);
  a=gldote_GaussDiag(x,e,thetaGaussDiag);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaGaussDiag, &l_GaussDiag);
  cout << gl.transpose()<<endl;

  ArrayXd thetaGaussLinSD(2); thetaGaussLinSD <<1,10; // is the range
  cout << "Gaussian with SDs linear between "<<thetaGaussLinSD(0)<<" and "<<thetaGaussLinSD(1)<<endl;
  l=l_GaussLinSD(x,thetaGaussLinSD);
  gl=gl_GaussLinSD(x,thetaGaussLinSD);
  a=gldote_GaussLinSD(x,e,thetaGaussLinSD);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaGaussLinSD, &l_GaussLinSD);
  cout << gl.transpose()<<endl;

  ArrayXd thetaGaussLinH(2); thetaGaussLinH <<1,10; // is the range
  cout << "Gaussian with diag(H) linear between "<<thetaGaussLinH(0)<<" and "<<thetaGaussLinH(1)<<endl;
  l=l_GaussLinH(x,thetaGaussLinH);
  gl=gl_GaussLinH(x,thetaGaussLinH);
  a=gldote_GaussLinH(x,e,thetaGaussLinH);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  cout << "Numerical differentiation.\n Gradient:";
  gl=glp_from_lp(x, thetaGaussLinH, &l_GaussLinH);
  cout << gl.transpose()<<endl<<"Hessian:"<<endl;
  MatrixXd H=Hlp_from_lp(x,thetaGaussLinH,&l_GaussLinH);
  cout <<H<<endl;

  ArrayXd thetaPowLinSc(3); thetaPowLinSc <<4,1,10; // (pow,lo,hi)
  cout << "Exponential power "<<thetaPowLinSc(0)<<" with scales linear between "<<thetaPowLinSc(1)<<" and "<<thetaPowLinSc(2)<<endl;
  l=l_PowLinSc(x,thetaPowLinSc);
  gl=gl_PowLinSc(x,thetaPowLinSc);
  a=gldote_PowLinSc(x,e,thetaPowLinSc);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaPowLinSc, &l_PowLinSc);
  cout << gl.transpose()<<endl;

  ArrayXd thetaPowProdLinSc(3); thetaPowProdLinSc <<4,1,10; // (pow,lo,hi)
  cout << "Exponential power "<<thetaPowProdLinSc(0)<<" product, with scales linear between "<<thetaPowProdLinSc(1)<<" and "<<thetaPowProdLinSc(2)<<endl;
  l=l_PowProdLinSc(x,thetaPowProdLinSc);
  gl=gl_PowProdLinSc(x,thetaPowProdLinSc);
  a=gldote_PowProdLinSc(x,e,thetaPowProdLinSc);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaPowProdLinSc, &l_PowProdLinSc);
  cout << gl.transpose()<<endl;

  ArrayXd thetaLogis(2*d);
  thetaLogis.head(d)=ArrayXd::LinSpaced(d,0,0); // Expectations
  thetaLogis.tail(d)=ArrayXd::LinSpaced(d,1,10); // Scale parameters
  cout << "Product of Logistics with [E1,..,,Ed,Sc1,...,Scd]=\n"<< thetaLogis.matrix().transpose()<<endl;
  l=l_LogisticCpts(x,thetaLogis);
  gl=gl_LogisticCpts(x,thetaLogis);
  cout << l <<endl<<gl.transpose()<<endl;
  gl=glp_from_lp(x, thetaLogis, &l_LogisticCpts);
  cout << gl.transpose()<<endl;


  ArrayXd thetaMVtDiagLinSc(3); thetaMVtDiagLinSc <<10,1,10; // is the range
  cout << "MV Student-t with df="<<thetaMVtDiagLinSc(0)<<" and scalings linear between "<<thetaMVtDiagLinSc(1)<<" and "<<thetaMVtDiagLinSc(2)<<endl;
  l=l_MVtDiagLinSc(x,thetaMVtDiagLinSc);
  gl=gl_MVtDiagLinSc(x,thetaMVtDiagLinSc);
  a=gldote_MVtDiagLinSc(x,e,thetaMVtDiagLinSc);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetaMVtDiagLinSc, &l_MVtDiagLinSc);
  cout << gl.transpose()<<endl;
  
  ArrayXd thetatProdLinSc(3); thetatProdLinSc <<10,1,10; // is the range
  cout << "Product of Student-t with df="<<thetatProdLinSc(0)<<" and scalings linear between "<<thetatProdLinSc(1)<<" and "<<thetatProdLinSc(2)<<endl;
  l=l_tProdLinSc(x,thetatProdLinSc);
  gl=gl_tProdLinSc(x,thetatProdLinSc);
  a=gldote_tProdLinSc(x,e,thetatProdLinSc);
  cout << l <<endl<<gl.transpose()<<endl<<a<<endl;
  gl=glp_from_lp(x, thetatProdLinSc, &l_tProdLinSc);
  cout << gl.transpose()<<endl;

  ArrayXd thetaMVt(1+d*d);
  for (i=0;i<d;i++) { for (j=0;j<d;j++) {
      thetaMVt(0)=10; // df
      thetaMVt(1+i+d*j)=exp((double)abs(i-j)*log(0.5)); // inverse variance
    }}
  cout << "MV Student-t with df="<<thetaMVt(0)<<" and inverse variance matrix of:"<<endl;
  ArrayXd Vinvlong=thetaMVt.tail(d*d);
  Map<MatrixXd> Vinv(Vinvlong.data(),d,d);
  cout << Vinv<<endl;
  l=l_MVt(x,thetaMVt);
  cout << l<< endl; // checked in R that this is the right answer

  ArrayXd thetaSkewNormal(d+1);
  thetaSkewNormal(0)=2; // alpha
  thetaSkewNormal.tail(d)=ArrayXd::LinSpaced(d,1,10); // Scale parameters
  cout << "Product of SkewNormals with [alpha,Sc1,...,Scd]=\n"<< thetaSkewNormal.matrix().transpose()<<endl;
  l=l_SkewNormalCpts(x,thetaSkewNormal);
  gl=gl_SkewNormalCpts(x,thetaSkewNormal);
  cout << l <<endl<<gl.transpose()<<endl;
  gl=glp_from_lp(x, thetaSkewNormal, &l_SkewNormalCpts);
  cout << gl.transpose()<<endl;

  ArrayXd thetaQuartGauss(d+1);
  thetaQuartGauss(0)=3; // a
  thetaQuartGauss.tail(d)=ArrayXd::LinSpaced(d,1,10); // Scale parameters
  cout << "Product of Quartic and Gauss with [a,Sc1,...,Scd]=\n"<< thetaQuartGauss.matrix().transpose()<<endl;
  l=l_QuartGaussCpts(x,thetaQuartGauss);
  gl=gl_QuartGaussCpts(x,thetaQuartGauss);
  cout << l <<endl<<gl.transpose()<<endl;
  gl=glp_from_lp(x, thetaQuartGauss, &l_QuartGaussCpts);
  cout << gl.transpose()<<endl;

  ArrayXd thetaLogisticGauss(d+1);
  thetaLogisticGauss(0)=3; // a
  thetaLogisticGauss.tail(d)=ArrayXd::LinSpaced(d,1,10); // Scale parameters
  cout << "Product of Logistic and Gauss with [a,Sc1,...,Scd]=\n"<< thetaLogisticGauss.matrix().transpose()<<endl;
  l=l_LogisticGaussCpts(x,thetaLogisticGauss);
  gl=gl_LogisticGaussCpts(x,thetaLogisticGauss);
  cout << l <<endl<<gl.transpose()<<endl;
  gl=glp_from_lp(x, thetaLogisticGauss, &l_LogisticGaussCpts);
  cout << gl.transpose()<<endl;

  ArrayXd thetaBananaTwo(d+1);
  thetaBananaTwo(0)=.8; // lambda
  thetaBananaTwo.tail(d)=ArrayXd::LinSpaced(d,1,10); // Scale parameters
  cout << "BananaTwo [lambda, Scs]=\n"<< thetaBananaTwo.matrix().transpose()<<endl;
  l=l_BananaTwo(x,thetaBananaTwo);
  gl=gl_BananaTwo(x,thetaBananaTwo);
  cout << l <<endl<<gl.transpose()<<endl;
  gl=glp_from_lp(x, thetaBananaTwo, &l_BananaTwo);
  cout << gl.transpose()<<endl;

  ArrayXd thetaBimodalGauss(d+1);
  thetaBimodalGauss(0)=.8; // a
  thetaBimodalGauss.tail(d)=ArrayXd::LinSpaced(d,1,10); // Scale parameters
  cout << "BimodalGauss [lambda, Scs]=\n"<< thetaBimodalGauss.matrix().transpose()<<endl;
  l=l_BimodalGauss(x,thetaBimodalGauss);
  gl=gl_BimodalGauss(x,thetaBimodalGauss);
  cout << l <<endl<<gl.transpose()<<endl;
  gl=glp_from_lp(x, thetaBimodalGauss, &l_BimodalGauss);
  cout << gl.transpose()<<endl;

  ArrayXd thetaPlusPrism(d+1);
  thetaPlusPrism(0)=.8; // lam
  thetaPlusPrism.tail(d)=ArrayXd::LinSpaced(d,1,10); // Scale parameters
  cout << "PlusPrism [lambda, Scs]=\n"<< thetaPlusPrism.matrix().transpose()<<endl;
  l=l_PlusPrism(x,thetaPlusPrism);
  gl=gl_PlusPrism(x,thetaPlusPrism);
  cout << l <<endl<<gl.transpose()<<endl;
  gl=glp_from_lp(x, thetaPlusPrism, &l_PlusPrism);
  cout << gl.transpose()<<endl;


  
  // Check the ancillary functions at the boundaries
  ArrayXd ys=ArrayXd::LinSpaced(23,-11,11);
  ys(0)=-20; ys(22)=20;
  cout <<"\nPI="<<_PI<<endl;
  cout <<"xs="<<ys.matrix().transpose()<<endl;
  cout <<"logPhi xs"<<logPhi(ys).matrix().transpose()<<endl;
  cout <<"phi/Phi xs"<<phiOPhi(ys).matrix().transpose()<<endl;
  
  return 0;
}

*/
